package org.netkernel.layer0.util;

import org.w3c.dom.*;

import java.lang.reflect.*;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicBoolean;

import javax.xml.namespace.QName;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

/** Bridge to XPath functionality. Attempts to use fast xpath, then falls back (on JDK1.5)
 * to javax.xml.xpath, failing that to org.apache.xpath (from xalan)
 * @author Tony Butterfield
 *
 */
public class XPath
{
	private static javax.xml.xpath.XPath sXPath;
	private static int sCacheSize=20;
	private static Map<String,CompiledXPath> sCache = new ConcurrentHashMap<String,CompiledXPath>();
	
	private static class CompiledXPath
	{	public long mLRU;
		public XPathExpression mExpression;
		public AtomicBoolean mInUse = new AtomicBoolean(false);
		
		public CompiledXPath(XPathExpression aExpression)
		{	mExpression=aExpression;
			touch();
		}
		
		public void touch()
		{	mLRU = System.currentTimeMillis();
		}
	}
	
	private static Object eval(String aXPath, Node aNode, QName aType) throws XPathExpressionException
	{	CompiledXPath cxp = sCache.get(aXPath);
		if (cxp==null)
		{	if (sCache.size()>sCacheSize)
			{	synchronized(sCache)
				{	if (sCache.size()>sCacheSize)
					{	purge();
					}
				}
			}
			cxp = new CompiledXPath(sXPath.compile(aXPath));
			sCache.put(aXPath, cxp);
		}
		else
		{	cxp.touch();
		}
		
		boolean free=cxp.mInUse.compareAndSet(false, true);
		if (free)
		{	try
			{	return cxp.mExpression.evaluate(aNode, aType);
			}
			finally
			{	cxp.mInUse.set(false);
			}
		}
		else
		{	//just compile it again in this case
			return sXPath.compile(aXPath).evaluate(aNode, aType);
		}		
	}
	
	private static void purge()
	{	long oldest=Long.MAX_VALUE;
		long avgAge=0;
		long now=System.currentTimeMillis();
		int count=0;
		for (Iterator i=sCache.values().iterator(); i.hasNext(); )
		{	CompiledXPath cxp = (CompiledXPath)i.next();
			if (cxp.mLRU<oldest)
			{	oldest=cxp.mLRU;
			}
			long age=now-cxp.mLRU;
			avgAge+=age;
			count++;
		}
		avgAge/=(long)count;
		oldest=now-oldest;
		long threshold=now-(oldest+avgAge)/2;
		count=0;
		for (Iterator i=sCache.entrySet().iterator(); i.hasNext(); )
		{	Map.Entry e = (Map.Entry)i.next();
			CompiledXPath cxp=(CompiledXPath)e.getValue();
			long lru=cxp.mLRU;
			if (lru<threshold)
			{	i.remove();
				count++;
			}
		}
	}
	
	static
	{ init();
	}
	
	public static void init()
	{	try
		{	sXPath=XPathFactory.newInstance().newXPath();		
		}
		catch (Exception e)
		{	e.printStackTrace();
		}	
	}
	
	public static List eval(String aXPath, Node aContextNode) throws IllegalArgumentException, XPathExpressionException
	{	List result;
		if (FastXPath.isSuitable(aXPath))
		{	result = FastXPath.eval(aContextNode,aXPath);
		}
		else
		{	NodeList nl = (NodeList)eval(aXPath, aContextNode, XPathConstants.NODESET);
			int length=nl.getLength();
			result = new ArrayList(length);
			for (int i=0; i<length; i++)
			{	result.add(nl.item(i));
			}
		}
		return result;
	}
	
	static boolean isTrue(String aXPath, Node aContextNode) throws IllegalArgumentException, XPathExpressionException
	{	boolean result;
		if (FastXPath.isSuitable(aXPath))
		{	result = FastXPath.isTrue(aContextNode,aXPath);
		}
		else
		{	Boolean b = (Boolean)eval(aXPath, aContextNode,XPathConstants.BOOLEAN);
			result=b.booleanValue();
		}
		return result;
	}
	
	static Node getSingleNode(String aXPath, Node aContextNode) throws XPathExpressionException
	{	Node result;
		if (FastXPath.isSuitable(aXPath))
		{	result = FastXPath.getSingleNode(aContextNode,aXPath);
		}
		else
		{	result = (Node)eval(aXPath, aContextNode,XPathConstants.NODE);
		}
		return result;
	}

}
